from pathlib import Path
from collections import defaultdict
from collections.abc import Iterable

import numpy as np
import nibabel as nib

from .base import SpatialIndex
from .util import logger


class SpatialIndexWriteExc(Exception):
    pass


class WritableSpatialIndex(SpatialIndex):

    writable = True

    _buffer: dict[tuple[int, int, int], bytes] = defaultdict(bytes)

    def __init__(self, filepath):
        super().__init__(filepath, mode="w")
        self.closed = False

    def write(self, pos, data):
        if self.closed:
            raise SpatialIndexWriteExc("WritableIndex already closed.")
        pos = self.validate_pos(pos)
        assert isinstance(
            data, Iterable
        ), f"Expecting data to be instance of Iterable, but was not: data={data}"
        for p, datum in zip(pos.tolist(), data):
            assert isinstance(
                datum, bytes
            ), f"Expecting datum to be of type bytes, but was not: datum={datum}"
            self._buffer[tuple(p)] += datum

    def save(self, affine=np.eye(4), shape: tuple[int, int, int] = None):
        if self.closed:
            raise SpatialIndexWriteExc("WritableIndex already closed.")

        voxel_file = str(self.filepath) + self.VOXEL_SUFFIX
        attr_file = str(self.filepath) + self.ATTR_SUFFIX
        manifestfile_file = str(self.filepath)

        if Path(voxel_file).exists():
            raise SpatialIndexWriteExc(
                f"File {voxel_file} already exist. Flushing incomplete files not yet supported"
            )

        if Path(attr_file).exists():
            raise SpatialIndexWriteExc(
                f"File {attr_file} already exist. Flushing incomplete files not yet supported"
            )

        total_bytes = 0
        for buf in self._buffer.values():
            total_bytes += len(buf)

        if total_bytes > self.UINT32_MAX:
            raise SpatialIndexWriteExc(
                "Spatial index data > uint32max. Currently not supported"
            )

        X, Y, Z = np.array(list(self._buffer.keys())).T
        minshape = np.array([(np.max(X) + 1, np.max(Y) + 1, np.max(Z) + 1)])

        if shape:
            shape = np.array(shape)
            assert shape.shape == (
                3,
            ), f"Expected shape have of shape (3,), but got {shape.shape}"
            assert np.all(
                shape >= minshape
            ), f"Expecting provided shape {shape} >= minshape {minshape}"
        else:
            shape = minshape

        # shape needs to be 1 + max index
        arr = np.zeros(shape, dtype=np.uint64)
        _offset_counter = 0

        total_count = len(self._buffer)
        logger.debug(f"Total buffer items: {total_count}")
        with open(attr_file, "wb") as fp:

            for idx, (key, _towrite) in enumerate(self._buffer.items()):
                if idx % 10 == 0:
                    logger.debug(
                        f"Progress: {idx} / {total_count} (updated every 10 items)"
                    )

                # as amazing it appears, bitshift has lower priority than addition
                # e.g. python -c "print(1 << 1 + 1)" prints 4, rather than 3
                arr[key] = np.uint64((_offset_counter << 32) + len(_towrite))

                _offset_counter += len(_towrite)
                fp.write(_towrite)

        # needs to explicitly specify uint64 dtype
        img = nib.Nifti1Image(arr, affine=affine, dtype=np.uint64)
        nib.save(img, voxel_file)

        with open(manifestfile_file, "wb") as fp:
            fp.write(self.HEADER.encode("utf-8"))
        self.closed = True



from typing import ContextManager, Callable
import gzip
from io import BufferedReader

import numpy as np
import nibabel as nib

from .base import SpatialIndex, V0SpatialIndex


class ReadableSpatialIndex(SpatialIndex):

    readable = True

    def __init__(
        self,
        filepath,
        *,
        reader_read: Callable[[str], ContextManager[BufferedReader]] = lambda p: open(
            p, mode="rb"
        ),
    ):
        super().__init__(filepath, mode="r")

        self.reader_read = reader_read

        with self.reader_read(filepath) as fp:
            meta = fp.read().decode()

        if meta.startswith(V0SpatialIndex.HEADER):
            self.VOXEL_SUFFIX = V0SpatialIndex.VOXEL_SUFFIX
            self.ATTR_SUFFIX = V0SpatialIndex.ATTR_SUFFIX

        voxel_filepath = str(self.filepath) + self.VOXEL_SUFFIX
        with self.reader_read(voxel_filepath) as fp:
            voxel_file_bytes = fp.read(-1)
            try:
                voxel_file_bytes = gzip.decompress(voxel_file_bytes)
            except gzip.BadGzipFile:
                ...
            self._nii = nib.Nifti1Image.from_bytes(voxel_file_bytes)

        self.dataobj = np.array(self._nii.dataobj)
        assert (
            self._nii.get_data_dtype() == np.uint64
        ), f"Expected to be of type uint64, but was {self._nii.get_data_dtype()}"

    @property
    def affine(self):
        return self._nii.affine

    def read(self, pos):
        pos = self.validate_pos(pos)
        x, y, z = pos.T

        result = []

        with self.reader_read(str(self.filepath) + self.ATTR_SUFFIX) as reader:
            for val in self.dataobj[x, y, z].tolist():
                offset = val >> 32
                bytes_to_read = val & self.UINT32_MAX
                reader.seek(int(offset))
                decoded = reader.read(int(bytes_to_read))
                if len(decoded) == 0:
                    continue
                result.append(decoded)
            return result



import pytest
import json
from tempfile import TemporaryDirectory
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Union

import requests

from vlsi import ReadableSpatialIndex
from vlsi.base import V0SpatialIndex

url = "https://data-proxy.ebrains.eu/api/v1/buckets/reference-atlas-data/sparse-indices/colin27-jba30-hg"


@pytest.fixture(scope="session")
def local_si_name():
    name = "foo"

    def download(url: str, dest: Union[str, Path]):
        resp = requests.get(url)
        resp.raise_for_status()
        with open(dest, "wb") as fp:
            fp.write(resp.content)

    with TemporaryDirectory() as _dir:

        fname = Path(_dir) / name

        with ThreadPoolExecutor() as ex:
            list(
                ex.map(
                    download,
                    [
                        url,
                        url + V0SpatialIndex.VOXEL_SUFFIX,
                        url + V0SpatialIndex.ATTR_SUFFIX,
                    ],
                    [
                        fname,
                        fname.with_suffix(V0SpatialIndex.VOXEL_SUFFIX),
                        fname.with_suffix(V0SpatialIndex.ATTR_SUFFIX),
                    ],
                )
            )

        yield fname


@pytest.fixture(scope="session")
def local_spatial_index(local_si_name):
    yield ReadableSpatialIndex(str(local_si_name))


@pytest.fixture(scope="session")
def remote_spatial_index():

    class RemoteReader:
        def __init__(self, path: str):
            self.marker = 0
            self.session = requests.Session()
            self.url = path

        def read(self, bytes=-1):

            if bytes == -1:
                resp = self.session.get(self.url)
                resp.raise_for_status()
                return resp.content

            assert bytes > 0

            end = bytes + self.marker - 1
            headers = {"Range": f"bytes={self.marker}-{end}"}
            resp = self.session.get(self.url, headers=headers)
            resp.raise_for_status()
            return resp.content

        def seek(self, offset):
            self.marker = offset

    @contextmanager
    def reader_read(url: str):
        yield RemoteReader(url)

    yield ReadableSpatialIndex(url, reader_read=reader_read)


jba30_colin_args = [
    (
        [72, 142, 119],
        json.dumps(
            {
                "d35d2d": 0.0003129999968223274,
                "c4d406": 0.3191109895706177,
                "6652b2": 0.28016701340675354,
                "943ce3": 9.999999974752427e-07,
                "23e136": 0.0006760000251233578,
                "749a1e": 0.3054789900779724,
                "ef6127": 0.0942310020327568,
                "1f3c97": 2.099999983329326e-05,
            }
        ).encode("utf-8")
        + b"\n",
    ),
    (
        [136, 216, 111],
        json.dumps(
            {"a72203": 0.3991990089416504, "ab1fe3": 0.08381400257349014}
        ).encode("utf-8")
        + b"\n",
    ),
]


@pytest.mark.parametrize(
    "pos, expected",
    [
        *[
            pytest.param([p], [ex], id=f"single-pos-idx-{idx}")
            for idx, (p, ex) in enumerate(jba30_colin_args)
        ],
        pytest.param(
            [arg[0] for arg in jba30_colin_args],
            [arg[1] for arg in jba30_colin_args],
            id="merged-array",
        ),
    ],
)
def test_local_legacy_spatial_indices(
    pos, expected, local_spatial_index: ReadableSpatialIndex
):
    assert local_spatial_index.read(pos) == expected


@pytest.mark.parametrize(
    "pos, expected",
    [
        *[
            pytest.param([p], [ex], id=f"single-pos-idx-{idx}")
            for idx, (p, ex) in enumerate(jba30_colin_args)
        ],
        pytest.param(
            [arg[0] for arg in jba30_colin_args],
            [arg[1] for arg in jba30_colin_args],
            id="merged-array",
        ),
    ],
)
def test_remote_legacy_spatial_indices(
    pos, expected, remote_spatial_index: ReadableSpatialIndex
):
    assert remote_spatial_index.read(pos) == expected



import numpy as np
import pytest
from vlsi import ReadableSpatialIndex

validate_pos = ReadableSpatialIndex.validate_pos


# Parametrized test cases for valid inputs
@pytest.mark.parametrize(
    "input_pos, expected_shape",
    [
        (np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64), (2, 3)),
        (np.array([[1, 2, 3]]), (1, 3)),
        ([[9, 10, 11], [12, 13, 14]], (2, 3)),
        (np.array([[1, 2, 3], [4, 5, 6]], dtype=int), (2, 3)),
    ],
)
def test_valid_inputs(input_pos, expected_shape):
    """Test that valid inputs are correctly processed and cast to uint32."""
    result = validate_pos(input_pos)
    assert result.shape == expected_shape
    assert result.dtype == np.uint32


@pytest.mark.parametrize(
    "input_pos",
    [
        np.array([1, 2, 3]),  # 1D array
        np.array([[[1, 2]], [[3, 4]], [[5, 6]]]),  # 3D array
        np.array([[1, 2], [3, 4]]),  # Wrong column size
        np.array([[]]),  # Empty array
    ],
)
def test_invalid_shapes(input_pos):
    """Test that invalid shapes raise AssertionError."""
    with pytest.raises(AssertionError) as exc_info:
        validate_pos(input_pos)
    # Verify the error messages are informative
    error_msg = str(exc_info.value)
    assert "len(pos.shape) to be 2" in error_msg or "expecting Nx3 array" in error_msg


def test_large_values():
    """Test handling of large values near uint32 limit."""
    max_value = 2**31 - 1
    input_pos = np.array([[max_value, max_value - 1, max_value - 2]])
    result = validate_pos(input_pos)

    assert result.shape == (1, 3)
    assert result.dtype == np.uint32
    assert np.array_equal(result, input_pos)


def test_float_values():
    """Test conversion of float values to uint32."""
    input_pos = np.array([[1.5, 2.7, 3.8]], dtype=np.float64)
    result = validate_pos(input_pos)

    # Test that float values were properly cast to uint32 (truncated or floored)
    assert result.shape == (1, 3)
    assert result.dtype == np.uint32
    assert result[0, 0] == 1  # Should be 1 after truncation



import logging

SEP = b"\x00"


def encode_data(bufs: list[bytes]):
    """
    Encode a list of byte buffers into a single byte string.

    Parameters
    ----------
    bufs : list[bytes]
        List of byte buffers to encode.

    Returns
    -------
    bytes
        Encoded byte string with SEP separators between buffers and a trailing SEP.
    """
    return b"".join(encode_datum(b) for b in bufs)


def encode_datum(buf: bytes):
    assert SEP not in buf, f"Cannot encode \\x00, as it is used as a separator"
    return buf + SEP


def decode_data(buf: bytes):
    """
    Decode byte buffer into list of data segments.

    Parameters
    ----------
    buf : bytes
        Byte buffer to decode, must end with SEP character.

    Returns
    -------
    list
        List of byte segments (b'...' objects) from the buffer.

    Raises
    ------
    AssertionError
        If buffer fails specification

    Examples
    --------
    >>> data = b'\\x00abc\\x00def\\x00'
    >>> decode_data(data)
    [b'', b'abc', b'def']
    """
    buf = buf.removesuffix(SEP)
    return buf.split(SEP)


logger = logging.getLogger(__name__)



import pytest
from vlsi.util import encode_data, encode_datum, decode_data


# ============================================================================
# encode_data Tests
# ============================================================================


@pytest.mark.parametrize(
    "bufs,expected",
    [
        ([b"abc", b"def", b"ghi"], b"abc\x00def\x00ghi\x00"),
        ([b"hello"], b"hello\x00"),
        ([], b""),
        ([b"", b"abc", b""], b"\x00abc\x00\x00"),
        ([b"\x01\x02\x03", b"\xff\xfe"], b"\x01\x02\x03\x00\xff\xfe\x00"),
    ],
    ids=[
        "multiple_buffers",
        "single_buffer",
        "empty_list",
        "empty_buffers",
        "binary_data",
    ],
)
def test_encode_data(bufs, expected):
    result = encode_data(bufs)
    assert result == expected


# ============================================================================
# encode_datum Tests
# ============================================================================


@pytest.mark.parametrize(
    "buf,expected",
    [
        (b"hello", b"hello\x00"),
        (b"", b"\x00"),
    ],
    ids=["simple_buffer", "empty_buffer"],
)
def test_encode_datum(buf, expected):
    result = encode_datum(buf)
    assert result == expected


@pytest.mark.parametrize("buf", [b"\x00\x01\x02"], ids=["contain_illegal_char"])
def test_expect_fail(buf):
    with pytest.raises(AssertionError):
        encode_datum(buf)


# ============================================================================
# decode_data Tests
# ============================================================================


@pytest.mark.parametrize(
    "buf,expected",
    [
        (b"abc\x00def\x00ghi\x00", [b"abc", b"def", b"ghi"]),
        (b"hello\x00", [b"hello"]),
        (b"\x00abc\x00\x00", [b"", b"abc", b""]),
        (b"\x00", [b""]),
        (b"\x01\x02\x03\x00\xff\xfe\x00", [b"\x01\x02\x03", b"\xff\xfe"]),
    ],
    ids=[
        "multiple_segments",
        "single_segment",
        "empty_segments",
        "empty_data",
        "binary_data",
    ],
)
def test_decode_data(buf, expected):
    result = decode_data(buf)
    assert result == expected


# ============================================================================
# Round-Trip Tests
# ============================================================================


@pytest.mark.parametrize(
    "data",
    [
        [b"abc", b"def", b"ghi"],
        [b"", b"", b""],
        [b"hello", b"world"],
        [b""],
    ],
    ids=[
        "multiple_buffers",
        "empty_buffers",
        "text_data",
        "single_empty",
    ],
)
def test_encode_decode_round_trip(data):
    encoded = encode_data(data)
    decoded = decode_data(encoded)
    assert decoded == data


@pytest.mark.parametrize(
    "data",
    [
        [b"\x00\x01\x02", b"\xff\xfe\xfd"],
    ],
    ids=["illegal_characters"],
)
def test_encode_decode_round_trip_fail(data):
    with pytest.raises(AssertionError):
        encode_data(data)


@pytest.mark.parametrize(
    "datum",
    [
        b"hello",
        b"",
        b"\xff\xfe",
    ],
    ids=["text", "empty", "high_bytes"],
)
def test_encode_datum_decode_round_trip(datum):
    encoded = encode_datum(datum)
    decoded = decode_data(encoded)
    assert decoded == [datum]


@pytest.mark.parametrize(
    "datum",
    [
        b"\x00\x01\x02",
    ],
    ids=["illegal_char"],
)
def test_encode_datum_decode_round_trip_fail(datum):
    with pytest.raises(AssertionError):
        encode_datum(datum)
